message = {
    "ok": {
        "code": "OK"
        , "message": "No errors encountered"
        , "level": "INFO"
    },
    "not_implemented": {
        "code": "EMI-GEN-000"
        , "message": "Function not yet implemented. Please contact the development team and mention code 'EMI-GEN-000'."
        , "level": "ERROR"
    },
    "not_found": {
        "code": "EMI-GEN-001"
        , "message": "Could not find what we were looking for"
        , "level": "ERROR"
    },
    "os_error": {
        "code": "EMI-GEN-002"
        , "message": "An OS error occurred"
        , "level": "ERROR"
    },
    "ignore": {
        "code": "EMI-GEN-003"
        , "message": "Entry has been ignored"
        , "level": "DEBUG"
    },
    "undetermined": {
        "code": "EMI-GEN-004"
        ,
        "message": "Undetermined result. This should not happen. Please contact the development team and mention code 'EMI-GEN-004'."
        , "level": "ERROR"
    },
    "main_config_not_found": {
        "code": "EMI-GEN-005"
        , "message": "Main configuration file not found."
        , "level": "FATAL"
    },
    "jinja_template_name_not_provided": {
        "code": "EMI-JIN-001"
        , "message": "Internal error: Jinja template name not provided"
        , "level": "ERROR"
    },
    "jinja_template_not_found": {
        "code": "EMI-JIN-002"
        , "message": "Jinja template not found."
        , "level": "ERROR"
    }
}
